/* dwiparse.c
   Functions for reading DWI step charts

Copyright 2004 Damian Yerrick

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, provided that the above copyright notice(s) and this permission notice appear in all copies of the Software and that both the above copyright notice(s) and this permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not be used in advertising or otherwise to promote the sale, use or other dealings in this Software without prior written authorization of the copyright holder.

*/


#include <string.h>
#include <ctype.h>
#include "dwiparse.h"

#define DISPLAY_WHILE_PARSING 0



/* cppfgetc() **************
   Like fgetc() but removes C++ comments from the input stream.
*/
int cppfgetc(FILE *fp)
{
  int first_c = fgetc(fp);

  if(first_c != '/')
    return first_c;
  else  /* we may have started a comment */
  {
    int second_c = fgetc(fp);

    if(second_c != '/')  /* not a comment */
    {
      ungetc(second_c, fp);
      return first_c;
    }

    /* it's a comment.  clear out everything up to newline */
    do {
      second_c = fgetc(fp);
    } while(second_c != '\r' && second_c != '\n');
    return second_c;
  }
}



/* DWI Tags

#TITLE:x;               Set title to x
#ARTIST:x;              Set artist to x
#GAP:x;                 Add stop at row 0 of (x/1000) seconds
#BPM:x;                 Add BPM change at row 0 to x BPM
#DISPLAYTITLE:x;        Ignore for now
#DISPLAYARTIST:x;       Ignore for now
#DISPLAYBPM:*;          Do #DISPLAYBPM:0..0;  (will show up random)
#DISPLAYBPM:a;          Do #DISPLAYBPM:a..a;
#DISPLAYBPM:a..b;       Set min_bpm to min(a, b) and max_bpm to max(a, b)
#FREEZE:a=b{,a=b}*;     Add freeze at row a*12 of b seconds
#CHANGEBPM:a=b{,a=b}*;  Add BPM change at row a*12 to b BPM
#COUPLE:d:f:s;          Add dance-couple stepchart with
                        difficulty class d, difficulty meter f, and steps s
#DOUBLE:d:f:s;          Add dance-double stepchart with
                        difficulty class d, difficulty meter f, and steps s
#SINGLE:d:f:s;          Add dance-single stepchart with
                        difficulty class d, difficulty meter f, and steps s
#SOLO:d:f:s;            Add dance-solo stepchart with
                        difficulty class d, difficulty meter f, and steps s
*/

/* Difficulty Classes

0: beginner
1: basic, light
2: another, trick, standard
3: maniac, ssr, heavy
4: smaniac, oni, challenge
*/


/* dwi_gets() **************
   Reads characters from infile until an end or EOF character is
   reached, whichever comes first, and stores the first (num - 1)
   characters in string.  Unlike fgets(), dwi_gets() does NOT
   include end in the read file and DISCARDS everything after
   (num - 1) characters.
*/
static char *dwi_gets(char *string, int num, int end, FILE *infile)
{
  int c;
  int len = 0;

  if(feof(infile) || ferror(infile))
    return NULL;
  for(c = cppfgetc(infile);
      c != EOF && c != end;
      c = cppfgetc(infile))
  {
    if(len < num - 1)
      string[len++] = c;
  }
  string[len] = 0;
  return string;
}


/* ignore_dwi_tag_content() **************
   For debugging, ignores the content of a tag in DWI.
   Pre-condition: after initial :
   Post-condition: after final ;
   Does NOT work with #BACKGROUND.
*/
static int ignore_dwi_tag_content(FILE *infile)
{
  int c;

  for(c = cppfgetc(infile);
      c != EOF && c != ';';
      c = cppfgetc(infile))
  {

  }
  return 0;
}


/* dwi_get2doubles() *****************
   Gets a pair of values of the form '1.2=3.4' from a DWI stream.
   Returns ',' if more follow, ';' if no more follow, or 0 for error.
   On error does not modify *a or *b.
*/
static int dwi_get2doubles(FILE *infile, double *a, double *b)
{
  int c;
  int len = 0;
  char string[64];
  char *endstr;
  double val1, val2;

  if(feof(infile) || ferror(infile))
    return 0;
  for(c = cppfgetc(infile);
      c != EOF && c != ';' && c != ',';
      c = cppfgetc(infile))
  {
    if(len < sizeof(string) - 1)
      string[len++] = c;
  }
  if(c != ',')
    c = ';';
  string[len] = 0;

  val1 = strtod(string, &endstr);
  if(endstr[0] != '=')
    return 0;
  val2 = strtod(endstr + 1, &endstr);

  if(a)
    *a = val1;
  if(b)
    *b = val2;
  return c;
}


static int dwi_title(KADParsedSong *song, FILE *infile)
{
  if(!song)
    return -1;
  dwi_gets(song->title, sizeof(song->title), ';', infile);
#if DISPLAY_WHILE_PARSING
  printf("Song title: %s\n", song->title);
#endif
  return 0;
}


static int dwi_artist(KADParsedSong *song, FILE *infile)
{
  dwi_gets(song->artist, sizeof(song->artist), ';', infile);
#if DISPLAY_WHILE_PARSING
  printf("Featured performer: %s\n", song->artist);
#endif
  return 0;
}


static int dwi_bpm(KADParsedSong *song, FILE *infile)
{
  char title[64] = "";
  double bpm;
  char *endbpm;

  dwi_gets(title, sizeof(title), ';', infile);
  bpm = strtod(title, &endbpm);

#if DISPLAY_WHILE_PARSING
  printf("Tempo change: 0:0 = %f BPM\n", bpm);
#endif
  KADSync_AddBPM(song->sync, 0, KADREAL(bpm));
  return 0;
}


static int dwi_displaybpm(KADParsedSong *song, FILE *infile)
{
  char title[64] = "";
  double bpm1, bpm2;
  char *endbpm;

  dwi_gets(title, sizeof(title), ';', infile);
  if(title[0] == '*')
  {
    bpm1 = bpm2 = 0;
  }
  else
  {
    bpm1 = strtod(title, &endbpm);
    if(endbpm[0] == '.')
    {
      while(endbpm[0] == '.')
        endbpm++;
      bpm2 = strtod(title, &endbpm);
      if(bpm2 < bpm1)
      {
        double t = bpm2; bpm2 = bpm1; bpm1 = t;
      }
    }
    else
    {
      bpm2 = bpm1;
    }
  }

  song->sync->min_disp_speed = KADFIX(bpm1);
  song->sync->max_disp_speed = KADFIX(bpm2);

#if DISPLAY_WHILE_PARSING
  if(bpm1 > 0 && bpm2 > 0)
    printf("Force tempo in songwheel: %f - %f BPM\n", 
            bpm1, bpm2);
  else
    puts("Force haywire tempo in songwheel");
#endif
  return 0;
}


static int dwi_changebpm(KADParsedSong *song, FILE *infile)
{
  int c;

  do {
    double a, b;

    c = dwi_get2doubles(infile, &a, &b);
    if(c)
    {
      unsigned int row_n = (int)(12.0 * a);
#if DISPLAY_WHILE_PARSING
      printf("Tempo change: %u:%u = %f BPM\n",
             row_n / 192, row_n % 192, b);
#endif
      KADSync_AddBPM(song->sync, row_n, KADREAL(b));
    }
  } while(c == ',');
  return 0;
}


static int dwi_gap(KADParsedSong *song, FILE *infile)
{
  char title[64] = "";
  double gap;
  char *endgap;

  dwi_gets(title, sizeof(title), ';', infile);
  gap = strtod(title, &endgap);

#if DISPLAY_WHILE_PARSING
  printf("Stop: 0:0 = %f seconds\n", gap * .001);
#endif
  KADSync_AddStop(song->sync, 0, KADREAL(gap / 1000.));
  return 0;
}


static int dwi_samplestart(KADParsedSong *song, FILE *infile)
{
  char title[64] = "";
  double gap;
  char *endgap;

  dwi_gets(title, sizeof(title), ';', infile);
  gap = strtod(title, &endgap);

#if DISPLAY_WHILE_PARSING
  printf("Songwheel sample start: %f seconds\n", gap);
#endif
  song->sync->sample_start = KADFIX(gap);
  return 0;
}


static int dwi_samplelength(KADParsedSong *song, FILE *infile)
{
  char title[64] = "";
  double gap;
  char *endgap;

  dwi_gets(title, sizeof(title), ';', infile);
  gap = strtod(title, &endgap);

#if DISPLAY_WHILE_PARSING
  printf("Songwheel sample plays for %f seconds\n", gap);
#endif
  song->sync->sample_duration = KADFIX(gap);
  return 0;
}


static int dwi_freeze(KADParsedSong *song, FILE *infile)
{
  int c;

  do {
    double a, b;

    c = dwi_get2doubles(infile, &a, &b);
    if(c)
    {
      unsigned int row_n = (int)(12.0 * a);
#if DISPLAY_WHILE_PARSING
      printf("Stop: %u:%u = %f seconds\n",
             row_n / 192, row_n % 192, b * .001);
#endif
      KADSync_AddStop(song->sync, row_n, KADREAL(b / 1000.));
    }
  } while(c == ',');
  return 0;
}


static const char dwi_difficulty_titles[6][16] =
{
  "EDIT", "BEGINNER", "BASIC", "ANOTHER", "MANIAC", "SMANIAC"
};


static int dwi_do_difficulty(KADParsedSong *song, FILE *infile, int *difftype_dst)
{
  char title[64] = "";
  int dif;

  dwi_gets(title, sizeof(title), ':', infile);
#if DISPLAY_WHILE_PARSING
  fputs(title, stdout);
#endif

  if(difftype_dst)
  {
    *difftype_dst = 0;
    for(dif = 0; dif < 6; dif++)
    {
      if(!strcmp(title, dwi_difficulty_titles[dif]))
      *difftype_dst = dif;
    }
  }

  dwi_gets(title, sizeof(title), ':', infile);
  dif = strtol(title, NULL, 10);
#if DISPLAY_WHILE_PARSING
  printf(" (%d')\n", dif);
#endif
  return dif;
}


/* KAD column numbers: 
   1 = 1P left, 2 = 1P down, 3 = up, 4 = right,
   5 = 2P left or 1P upleft, 6 = 2P down, 7 = 2P up, 8 = 2P right or 1P upright
*/

static const char dwi_number_values[10][2] =
{
  {0, 0}, {1, 2}, {0, 2}, {2, 3}, {1, 0},  /* 0-4 */
  {0, 0}, {4, 0}, {1, 3}, {0, 3}, {4, 3}   /* 5-9 */
};

static const char dwi_letter_values[13][2] =
{
  {2, 3}, {1, 4}, {5, 0}, {8, 0}, {1, 5},  /* A-E */
  {2, 5}, {3, 5}, {4, 5}, {1, 8}, {2, 8},  /* F-J */
  {3, 8}, {4, 8}, {5, 8}                   /* K-M */
};

/* DWI density control commands:

default 1/8
(       1/16
[       1/24
{       1/64
`       1/192
)]}'    1/8
<       set in_jump
>       clear in_jump and move one down
*/


static void dwi_tie_step(KADParsedRow *steps, 
                         size_t start, size_t end, unsigned int col)
{
  for(; start < end; start++)
  {
    if(steps[start].column == col)
      steps[start].column = col | 0x80;
  }
}

static int dwi_read_chart(KADChart *chart, FILE *infile)
{
  int c;
  unsigned char in_jump = 0, in_bang = 0;
  int cur_measure = 0;
  unsigned char cur_row = 0;
  unsigned char increment = 192/8;
  unsigned int row_start = 0, prev_row_start = 0;

  for(c = cppfgetc(infile);
      c != EOF && c != ':' && c != ';';
      c = cppfgetc(infile))
  {
    char col1 = 0, col2 = 0, advance = 0;

    if(c >= '0' && c <= '9')
    {
      col1 = dwi_number_values[c - '0'][0];
      col2 = dwi_number_values[c - '0'][1];
      advance = !in_jump;
    }
    else if(c >= 'A' && c <= 'M')
    {
      col1 = dwi_letter_values[c - 'A'][0];
      col2 = dwi_letter_values[c - 'A'][1];
      advance = !in_jump;
    }
    else if(c == '>')
      advance = 1;
    else if(c == ')' || c == ']' || c == '}' || c == '\'')
      increment = 192/8;
    else if(c == '(')
      increment = 192/16;
    else if(c == '[')
      increment = 192/24;
    else if(c == '{')
      increment = 192/64;
    else if(c == '`')
      increment = 192/192;
    else if(c == '!')
      in_bang = 1;
      

    if(col1)
    {
      if(in_bang)
        dwi_tie_step(chart->steps->data, prev_row_start, chart->steps->num, col1);
      else
      {
        if(KADChart_AddRow(chart, cur_measure, cur_row, col1) < 0)
          return -1;
      }
    }
    if(col2)
    {
      if(in_bang)
        dwi_tie_step(chart->steps->data, prev_row_start, chart->steps->num, col2);
      else
      {
        if(KADChart_AddRow(chart, cur_measure, cur_row, col2) < 0)
          return -1;
      }
    }

    if(advance)
    {
      if(in_bang)
      {
        in_bang = 0;
      }
      else
      {
        prev_row_start = row_start;
        row_start = chart->steps->num;
        cur_row += increment;
        if(cur_row >= 192)
        {
          cur_row -= 192;
          cur_measure++;
        }
      }
    }
  }

  return 0;
}


static int dwi_single(KADParsedSong *song, FILE *infile)
{
  KADChart *c;
  int dif, diftype;

#if DISPLAY_WHILE_PARSING
  fputs("Single (4 panel) stepchart: ", stdout);
#endif
  dif = dwi_do_difficulty(song, infile, &diftype);
  c = KADChart_new(dif);
  if(!c)
    return -1;
  if(dwi_read_chart(c, infile) < 0)
  {
    KADChart_delete(c);
    return -1;
  }
  c->steps_type = 'S';
  c->difficulty_type = diftype;
  dif = KADParsedSong_AddChart(song, c);
  if(dif < 0)
  {
    KADChart_delete(c);
    return -1;
  }
#if DISPLAY_WHILE_PARSING
  dwi_print_chart(c);
#endif
  return 0;
}


static int dwi_solo(KADParsedSong *song, FILE *infile)
{
  KADChart *c;
  int dif, diftype;

#if DISPLAY_WHILE_PARSING
  fputs("Solo (6 panel) stepchart: ", stdout);
#endif
  dif = dwi_do_difficulty(song, infile, &diftype);
  c = KADChart_new(dif);
  if(!c)
    return -1;
  if(dwi_read_chart(c, infile) < 0)
  {
    KADChart_delete(c);
    return -1;
  }
  c->steps_type = '6';
  c->difficulty_type = diftype;
  dif = KADParsedSong_AddChart(song, c);
  if(dif < 0)
  {
    KADChart_delete(c);
    return -1;
  }
#if DISPLAY_WHILE_PARSING
  dwi_print_chart(c);
#endif
  return 0;
}


static int dwi_double(KADParsedSong *song, FILE *infile)
{
  KADChart *lside;
  KADChart *rside;
  KADChart *merged;
  int dif, diftype;

#if DISPLAY_WHILE_PARSING
  fputs("Double (8 panel) stepchart: ", stdout);
#endif
  dif = dwi_do_difficulty(song, infile, &diftype);

  lside = KADChart_new(dif);
  rside = KADChart_new(dif);
  if(!lside || !rside)
  {
    if(lside)
      KADChart_delete(lside);
    if(rside)
      KADChart_delete(rside);
    return -1;
  }

  if(dwi_read_chart(lside, infile) < 0 ||
     dwi_read_chart(rside, infile) < 0)
  {
    KADChart_delete(lside);
    KADChart_delete(rside);
    return -1;
  }

  merged = KADChart_merge(lside, rside, 4);
  if(!merged)
  {
    KADChart_delete(lside);
    KADChart_delete(rside);
    return -1;
  }
  KADChart_delete(lside);
  KADChart_delete(rside);

  merged->steps_type = 'D';
  merged->difficulty_type = diftype;
  dif = KADParsedSong_AddChart(song, merged);
  if(dif < 0)
  {
    KADChart_delete(merged);
    return -1;
  }

#if DISPLAY_WHILE_PARSING
  dwi_print_chart(merged);
#endif
  return 0;
}


static int dwi_couple(KADParsedSong *song, FILE *infile)
{
  KADChart *lside;
  KADChart *rside;
  KADChart *merged;
  int dif, diftype;

#if DISPLAY_WHILE_PARSING
  fputs("Couple (8 panel 2 player) stepchart: ", stdout);
#endif
  dif = dwi_do_difficulty(song, infile, &diftype);

  lside = KADChart_new(dif);
  rside = KADChart_new(dif);
  if(!lside || !rside)
  {
    if(lside)
      KADChart_delete(lside);
    if(rside)
      KADChart_delete(rside);
    return -1;
  }

  if(dwi_read_chart(lside, infile) < 0 ||
     dwi_read_chart(rside, infile) < 0)
  {
    KADChart_delete(lside);
    KADChart_delete(rside);
    return -1;
  }
  puts("read L and R");
  merged = KADChart_merge(lside, rside, 4);
  if(!merged)
  {
    KADChart_delete(lside);
    KADChart_delete(rside);
    return -1;
  }
  KADChart_delete(lside);
  KADChart_delete(rside);

  merged->steps_type = 'C';
  merged->difficulty_type = diftype;
  dif = KADParsedSong_AddChart(song, merged);
  if(dif < 0)
  {
    KADChart_delete(merged);
    return -1;
  }

#if DISPLAY_WHILE_PARSING
  dwi_print_chart(merged);
#endif
  return 0;
}


typedef struct DWITagDispatch
{
  char name[16];
  int (*func)(KADParsedSong *, FILE *);
} DWITagDispatch;


/* DWI tags, must be listed in alphabetical order */
static const DWITagDispatch dwi_tags[] =
{
  {"ARTIST", dwi_artist},
  {"BPM", dwi_bpm},
  {"CHANGEBPM", dwi_changebpm},
  {"COUPLE", dwi_couple},
  {"DISPLAYBPM", dwi_displaybpm},
  {"DOUBLE", dwi_double},
  {"FREEZE", dwi_freeze},
  {"GAP", dwi_gap},
  {"SAMPLELENGTH", dwi_samplelength},
  {"SAMPLESTART", dwi_samplestart},
  {"SINGLE", dwi_single},
  {"SOLO", dwi_solo},
  {"TITLE", dwi_title}
};


/* get_dwi_tag() **************
   Interprets a tag in DWI.
   Pre-condition: after initial #
   Post-condition: after final ;
*/
static void get_dwi_tag(KADParsedSong *song, FILE *infile)
{
  char got_kw[16] = "";
  size_t first = 0;
  size_t last = (sizeof(dwi_tags) / sizeof(dwi_tags[0])) - 1;

#if 0
  for(kw_len = 0; kw_len < sizeof(got_kw) - 1; )
  {
    int c = cppfgetc(infile);

    if(c == ':')
      break;
    else if(isupper(c))
      got_kw[kw_len++] = c;
  }
  got_kw[kw_len] = 0;
#else
  dwi_gets(got_kw, sizeof(got_kw), ':', infile);
#endif

#if 0
  printf("DWI keyword #%s:\n", got_kw);
#endif
  /* binary search for dispatch */
  while(first < last)
  {
    size_t mid = (first + last) / 2;
    int result = strcmp(got_kw, dwi_tags[mid].name);

#if 0
    printf("%u, %u, %u, %s, %s\n", 
           (unsigned int)first, (unsigned int)mid, (unsigned int)last,
           got_kw, dwi_tags[mid].name);
#endif
    if(result < 0)  /* first half */
    {
      last = mid - (mid > first);
    }
    else if(result > 0)
    {
      first = mid + (mid < last);
    }
    else
      last = first = mid;
  }

#if 0
  printf("bsearch idx = %u\n", (unsigned int)first);
#endif
  if(strcmp(dwi_tags[first].name, got_kw))
  {
    printf("Skipping DWI keyword #%s:\n", got_kw);
    ignore_dwi_tag_content(infile);
  }
  else
    dwi_tags[first].func(song, infile);
}


/* KADParsedSong_ReadDWI() ***************
   Parses a DWI bytestream using get_dwi_tag().
*/
void KADParsedSong_ReadDWI(KADParsedSong *song, FILE *infile)
{
  int c;

  while((c = fgetc(infile)) != EOF)
  {
    if(c == '#')
      get_dwi_tag(song, infile);
  }
}
