/* kadchart.c
   Functions for handling step charts

Copyright 2004 Damian Yerrick

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, provided that the above copyright notice(s) and this permission notice appear in all copies of the Software and that both the above copyright notice(s) and this permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not be used in advertising or otherwise to promote the sale, use or other dealings in this Software without prior written authorization of the copyright holder.

*/

#include "kadchart.h"

/* KADV_expand() *********
   Expands a vector's memory to hold at least this many elements.
   Fails (returns negative) if it can't allocate more memory.
*/
int KADV_expand(KADV *c, size_t new_max)
{
  void *new_mem = realloc(c->data, new_max * c->size);
  if(!new_mem)
    return -1;
  c->max_num = new_max;
  c->data = new_mem;
  return 0;
}


/* KADV_growto() *****************
   Sets the size of a vector to n.
   Returns the old size.
*/
int KADV_growto(KADV *c, size_t n)
{
  int result;

  if(n > c->max_num)
  {
    result = KADV_expand(c, c->max_num + n);
    if(result < 0)
      return result;
  }
  result = c->num;
  c->num = n;
  return result;
}


/* KADV_growby() *****************
   Adds n to the size of the vector.
   Returns the old size.
*/
int KADV_growby(KADV *c, size_t n)
{
  return KADV_growto(c, n + c->num);
}

KADV *KADV_new(size_t size, size_t initial_max_num)
{
  KADV *c = malloc(sizeof(KADV));
  void *data = malloc(size * initial_max_num);

  if(!c || !data)
  {
    if(c) free(c);
    if(data) free(data);
    return NULL;
  }

  c->data = data;
  c->size = size;
  c->max_num = initial_max_num;
  c->num = 0;
  return c;
}

void KADV_delete(KADV *c)
{
  if(c)
  {
    if(c->data) free(c->data);
    free(c);
  }
}


/* KADChart_AddRow() ************
   Appends a row to a stepchart.
   Does nothing for null c.
   Guaranteed to succeed unless c->n_steps is already at c->max_steps;
   in that case, fails (returns negative) if it can't expand the chart memory.
*/
int KADChart_AddRow(KADChart *c, 
                    unsigned int measure, unsigned int row, unsigned int col)
{
  KADParsedRow *dst;

  if(!c || !c->steps || (KADV_growby(c->steps, 1) < 0))
    return -1;

  dst = (KADParsedRow *)c->steps->data + (c->steps->num - 1);
  dst->measure = measure;
  dst->row = row;
  dst->column = col;
  return 0;
}

#define KAD_NEW_CHART_SIZE 256

KADChart *KADChart_new(int difficulty)
{
  KADChart *c = malloc(sizeof(KADChart));
  KADV *steps = KADV_new(sizeof(KADParsedRow), KAD_NEW_CHART_SIZE);

  if(!c || !steps)
  {
    if(c) free(c);
    if(steps) KADV_delete(steps);
    return NULL;
  }

  c->steps = steps;
  c->difficulty = difficulty;
  return c;
}

void KADChart_delete(KADChart *c)
{
  if(c)
  {
    if(c->steps) KADV_delete(c->steps);
    free(c);
  }
}


/* KADChart_merge() ************
   Creates a new chart that merges the steps in two existing charts.
   Places chart b shift columns to the right.
   Used for double and couple.
   Returns NULL on failure.
*/
KADChart *KADChart_merge(const KADChart *a, const KADChart *b, int shift)
{
  size_t a_len, b_len;
  KADParsedRow *a_step, *b_step;
  KADChart *ch;

  if(!a || !b)
    return NULL;
  ch = KADChart_new(a->difficulty);
  if(!ch)
    return NULL;

  a_len = a->steps->num;
  b_len = b->steps->num;
  a_step = (KADParsedRow *)a->steps->data;
  b_step = (KADParsedRow *)b->steps->data;

  /* we know the number of steps */
  if(KADV_expand(ch->steps, a_len + b_len) < 0)
  {
    KADChart_delete(ch);
    return NULL;
  }

  while(a_len > 0 || b_len > 0)
  {
    if(a_len == 0)
    {
      KADChart_AddRow(ch, b_step->measure, b_step->row, b_step->column + shift);
      b_step++;
      b_len--;
    }
    else if(b_len == 0 || a_step->measure < b_step->measure)
    {
      KADChart_AddRow(ch, a_step->measure, a_step->row, a_step->column);
      a_step++;
      a_len--;
    }
    else if(a_step->measure > b_step->measure)
    {
      KADChart_AddRow(ch, b_step->measure, b_step->row, b_step->column + shift);
      b_step++;
      b_len--;
    }
    else if(a_step->row <= b_step->row)
    {
      KADChart_AddRow(ch, a_step->measure, a_step->row, a_step->column);
      a_step++;
      a_len--;
    }
    else
    {
      KADChart_AddRow(ch, b_step->measure, b_step->row, b_step->column + shift);
      b_step++;
      b_len--;
    }
  }
  return ch;
}


#define KAD_NEW_SYNC_BPMS 64   /* for songs with live drums */
#define KAD_NEW_SYNC_STOPS 16  /* more than this and you're Mooglehunter */

KADSync *KADSync_new(void)
{
  KADSync *c = malloc(sizeof(KADSync));
  KADV *bpms = KADV_new(sizeof(KADParsedRow), KAD_NEW_CHART_SIZE);
  KADV *stops = KADV_new(sizeof(KADParsedRow), KAD_NEW_CHART_SIZE);

  if(!c || !bpms || !stops)
  {
    if(c) free(c);
    if(bpms) KADV_delete(bpms);
    if(stops) KADV_delete(stops);
    return NULL;
  }

  c->bpms = bpms;
  c->stops = stops;
  c->min_real_speed = KADFIX(32767);
  c->max_real_speed = 0;
  c->min_disp_speed = KADFIX(32767);
  c->max_disp_speed = 0;
  c->sample_start = KADFIX(40);
  c->sample_duration = KADFIX(12);
  return c;
}


void KADSync_delete(KADSync *c)
{
  if(c)
  {
    if(c->bpms) KADV_delete(c->bpms);
    if(c->stops) KADV_delete(c->stops);
    free(c);
  }
}


int KADSync_AddBPM(KADSync *c, unsigned int row, KADReal speed)
{
  KADBPMChange *dst;

  if(!c || !c->bpms || (KADV_growby(c->bpms, 1) < 0))
    return -1;

  dst = (KADBPMChange *)c->bpms->data + (c->bpms->num - 1);
  dst->row = row;
  dst->bpm = speed;

  if(c->min_real_speed > speed)
    c->min_real_speed = speed;
  if(c->max_real_speed < speed)
    c->max_real_speed = speed;

  return 0;
}


int KADSync_AddStop(KADSync *c, unsigned int row, KADReal duration)
{
  KADStop *dst;

  if(!c || !c->stops || (KADV_growby(c->stops, 1) < 0))
    return -1;

  dst = (KADStop *)c->stops->data + (c->stops->num - 1);
  dst->row = row;
  dst->duration = duration;
  return 0;
}

int KADParsedSong_AddChart(KADParsedSong *song, KADChart *c)
{
  KADChart **dst;

  if(!song || !song->charts || !c
     || (KADV_growby(song->charts, 1) < 0))
    return -1;

  dst = (KADChart **)song->charts->data + (song->charts->num - 1);
  dst[0] = c;
  return 0;
}



KADParsedSong *KADParsedSong_new(void)
{
  KADParsedSong *c = malloc(sizeof(KADParsedSong));
  KADSync *sync = KADSync_new();
  KADV *charts = KADV_new(sizeof(KADChart *), KAD_NEW_CHART_SIZE);

  if(!c || !sync)
  {
    if(c) free(c);
    if(sync) KADSync_delete(sync);
    if(charts) KADV_delete(charts);
    return NULL;
  }

  c->title[0] = 0;
  c->artist[0] = 0;
  c->sync = sync;
  c->charts = charts;
  return c;
}

void KADParsedSong_delete(KADParsedSong *c)
{
  if(c)
  {
    if(c->sync) KADSync_delete(c->sync);
    if(c->charts)
    {
      unsigned int i = c->charts->num;
      KADChart **charts = c->charts->data;

      for(i = 0; i < c->charts->num; i--, charts++)
        if(*charts) free(*charts);
      KADV_delete(c->charts);
    }
    free(c);
  }
}

