/* sync_compile.c
   Find the start time for each BPM change

Copyright 2004 Damian Yerrick

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, provided that the above copyright notice(s) and this permission notice appear in all copies of the Software and that both the above copyright notice(s) and this permission notice appear in supporting documentation.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

Except as contained in this notice, the name of a copyright holder shall not be used in advertising or otherwise to promote the sale, use or other dealings in this Software without prior written authorization of the copyright holder.

*/

#include "kadchart.h"
#include <stdio.h>

/* The Game Boy Advance handheld game system's CPU runs at 16777216 Hz.
   It displays one frame of video every 280896 CPU cycles.  Thus, the
   refresh rate is somewhere around 59.7275 Hz.
*/
#define TICKS_PER_SECOND (16777216.0/280896.0)
#define TICKS_PER_MINUTE (60.0 * (TICKS_PER_SECOND))

/*
                  48 rows          1 minute
rows_tick = bpm * ------- * ----------------------
                  1 beat    TICKS_PER_MINUTE ticks

ticks_row = 1/rows_tick (corrected for bias)

for 300 BPM, rows_tick = 4.01825 and ticks_row = 0.24886
so store both values as 16.16 fixed

*/

typedef struct KADCompiledSync
{
  unsigned int row;
  KADFixed time;  /* starting time in ticks */
  KADFixed rows_tick;  /* speed in rows/tick; 0 denotes a freeze */
  KADFixed ticks_row;  /* period in ticks/row */
} KADCompiledSync;


static void KADCompiledSync_PrintRow(const KADCompiledSync *r)
{
  unsigned int measure = r->row / 192;
  unsigned int row = r->row - 192 * measure;
  unsigned int beat = row / 48;

  printf("%3u:%u:%02u at %6.2f seconds, ",
         measure + 1, beat + 1, row - 48 * beat,
         KADUNFIX(r->time) * (1.0 / TICKS_PER_SECOND));
  if(r->rows_tick)
  {
    printf("%f rows/tick = %.2f = %.2f BPM\n",
           KADUNFIX(r->rows_tick),
           TICKS_PER_MINUTE / 48.0 * KADUNFIX(r->rows_tick),
           TICKS_PER_MINUTE / 48.0 / KADUNFIX(r->ticks_row));
  }
  else
  {
    puts("Stopped");
  }
}


KADV *KADSync_compile(KADSync *sync)
{
  KADV *c = KADV_new(sizeof(KADCompiledSync), 64);
  const KADBPMChange *bpms = sync->bpms->data;
  const KADStop *stops = sync->stops->data;
  size_t n_bpms = sync->bpms->num;
  size_t n_stops = sync->stops->num;
  KADReal cur_time = 0;
  unsigned int cur_row = 0;

  if(!c)
    return 0;

  while(n_bpms > 1 || n_stops > 0)
  {
    unsigned int end_row = (n_bpms <= 1 || stops[0].row <= bpms[1].row)
                           ? stops[0].row
                           : bpms[1].row;

    if(n_stops > 0 && cur_row == end_row)  /* we're on a stop */
    {
      int idx = KADV_growby(c, 1);
      KADCompiledSync *dst = (idx >= 0) ? (KADCompiledSync *)c->data + idx : NULL;
      if(!dst)
      {
        KADV_delete(c);
        return 0;
      }

      dst->row = cur_row;
      dst->time = KADFIX(KADUNREAL(cur_time));
      dst->rows_tick = KADFIX(0);
      dst->ticks_row = KADFIX(32123);
      KADCompiledSync_PrintRow(dst);

      cur_time += TICKS_PER_SECOND * stops[0].duration;
      stops++;
      n_stops--;
    }
    else  /* we're on a segment of moving arrows */
    {
      double ticks_row = TICKS_PER_MINUTE / 48.0 / KADUNREAL(bpms[0].bpm);
      KADReal seg_time = KADREAL(ticks_row * (end_row - cur_row));
      int idx = KADV_growby(c, 1);
      KADCompiledSync *dst = (idx >= 0) ? (KADCompiledSync *)c->data + idx : NULL;
      if(!dst)
      {
        KADV_delete(c);
        return 0;
      }

      dst->row = cur_row;
      dst->time = KADFIX(KADUNREAL(cur_time));
      dst->rows_tick = KADFIX(1.0 / ticks_row);
      dst->ticks_row = KADFIX(ticks_row);
      KADCompiledSync_PrintRow(dst);

      cur_time += seg_time;
      cur_row = end_row;
      if(end_row >= bpms[1].row)
      {
        bpms++;
        n_bpms--;
      }
    }
  }
  if(n_bpms)
  {
    double ticks_row = TICKS_PER_MINUTE / 48.0 / KADUNREAL(bpms[0].bpm);
    int idx = KADV_growby(c, 1);
    KADCompiledSync *dst = (idx >= 0) ? (KADCompiledSync *)c->data + idx : NULL;
    if(!dst)
    {
      KADV_delete(c);
      return 0;
    }

    dst->row = cur_row;
    dst->time = KADFIX(KADUNREAL(cur_time));
    dst->rows_tick = KADFIX(1.0 / ticks_row);
    dst->ticks_row = KADFIX(ticks_row);
    KADCompiledSync_PrintRow(dst);
  }
  return c;
}

